open Dib_types
open Sock_io

(*
  To do:
  - Rationalize robot updates
*)

(****** Connection stuff ******)

(* send a command *)
let send_command cmd sock = send_ln sock (command_to_str cmd)

(* read board from socket and return it *)
let read_board sock = 
  let str = readline sock in
  let bases = ref [] in
  let (width, height) = Scanf.sscanf str "%d %d" (fun i j -> i,j) in
  let board = make_board width height in
  for i = 0 to width+1 do
    board.(i).(0) <- { kind=WALL;is_beach=false;has_robots=false};
    board.(i).(height+1) <- {kind=WALL;is_beach=false;has_robots=false};
  done;
  for y = 1 to height do
    board.(0).(y) <- {kind=WALL;is_beach=false;has_robots=false};
    board.(width+1).(y) <- {kind=WALL;is_beach=false;has_robots=false};
    let str = readline sock in
    let x = ref 1 in
    print_endline str;
    String.iter (fun ch ->
                   let sq = ch_square ch in
                   board.(!x).(y) <- {kind=sq;is_beach=false;has_robots=false};
                   begin
                     match sq with
                         BASE -> bases := (!x,y) :: !bases
                       | _ -> ()
                   end;
                   x := !x + 1)
      str
  done;
  for y = 1 to height do
    for x = 1 to width do
      if is_walkable board.(x).(y).kind then
        if is_walkable board.(x-1).(y).kind && is_water board.(x+1).(y).kind
        || is_walkable board.(x+1).(y).kind && is_water board.(x-1).(y).kind
        || is_walkable board.(x).(y-1).kind && is_water board.(x).(y+1).kind
        || is_walkable board.(x).(y+1).kind && is_water board.(x).(y-1).kind
        then
          let b = board.(x).(y) in
            board.(x).(y) <- {kind=b.kind; is_beach=true; has_robots=false};
    done
  done;
  board, bases

(* read my robot info line *)
let read_my_robot sock =
  let str = readline sock in
  Scanf.sscanf str "%d %d %d" (fun i j k -> i,j,k)

(* add all robots from lst to hash *)
let rec read_all_robot_fields hash lst =
  (* add robot with these attributes to hash *) 
  let read_robot_fields hash idstr f1 v1 f2 v2 =
    let read_pair f v rob =
      if f = "X" then (rob.x <- int_of_string v)
      else if f = "Y" then (rob.y <- int_of_string v) in
    let id = Str.string_after idstr 1 in
    let rob = new_robot id in
    read_pair f1 v1 rob;
    read_pair f2 v2 rob;
    Hashtbl.add hash id rob in
  match lst with
    | [] -> ()
    | id::f1::v1::f2::v2::rest ->
        read_robot_fields hash id f1 v1 f2 v2;
        read_all_robot_fields hash rest
    | _ -> raise Bad_input

let read_all_robots_str str =
  let fields = Str.split (Str.regexp " ") str in
  let robot_hash = Hashtbl.create 10 in
  read_all_robot_fields robot_hash fields;
  robot_hash

(* create hash and add all robots to it, returning it *)
let read_all_robots sock =
  let str = readline sock in
  read_all_robots_str str

(* add my robot's info to hash and return my robot *)
let merge_robot hash (myid, capacity, bankroll) =
  let rob = Hashtbl.find hash (string_of_int myid) in
  rob.capacity <- capacity;
  rob.bankroll <- bankroll;
  rob
  
(* Get a list of all packages at x y and return it *)
let get_package_info x y sock =
  let rec read_all_package_fields fields =
    match fields with
      | [] -> []
      | pkg_id::destx::desty::weight::rest ->
          { pkg_id=pkg_id;
            pkg_loc=Sitting(x,y);
            pkg_dest=Some(destx,desty);
            weight=Some(weight)
          } :: read_all_package_fields rest
      | x -> raise Bad_input in
  let str = readline sock in
  let fields = Str.split (Str.regexp " ") str in
  read_all_package_fields (list_map fields int_of_string)
    
let read_updates robots str =
  let rec read_robot_updates r updates =
    match updates with
        "N"::rest -> {robot=r; action=UpdateMove(N)} :: read_robot_updates r rest
      | "S"::rest -> {robot=r; action=UpdateMove(S)} :: read_robot_updates r rest
      | "E"::rest -> {robot=r; action=UpdateMove(E)} :: read_robot_updates r rest
      | "W"::rest -> {robot=r; action=UpdateMove(W)} :: read_robot_updates r rest
      | "P"::p::rest -> {robot=r; action=UpdatePick(int_of_string p)} :: read_robot_updates r rest
      | "D"::d::rest -> {robot=r; action=UpdateDrop(int_of_string d)} :: read_robot_updates r rest
      | "X"::x::rest -> {robot=r; action=UpdateX(int_of_string x)} :: read_robot_updates r rest
      | "Y"::y::rest -> {robot=r; action=UpdateY(int_of_string y)} :: read_robot_updates r rest
      | [] -> []
      |  _ -> raise Bad_input in
  match Str.split (Str.regexp " ") str with
      id::rest ->
        let r = (
          try Hashtbl.find robots id
          with Not_found ->
            let r = new_robot id in
            Hashtbl.add robots id r;
            r
        ) in
        r.alive <- true;
        read_robot_updates r rest
          | [] -> raise Bad_input

(* returns an update list *)
let receive_updates robots sock =
  let str = readline sock in
  let update_strs = Str.split (Str.regexp " ?#") str in
  let updates = list_map update_strs (fun s -> read_updates robots s) in
  List.flatten updates

let str_prefix_equal str pref =
  String.length str >= String.length pref
  && (String.sub str 0 (String.length pref)) = pref
;;

(* read a non-empty line; end of file -> "END" *)
let rec getline inp =
  try let str = input_line inp in
  if str_prefix_equal str "(*" then begin
    getline inp
  end else begin
    str
  end
  with End_of_file -> "END";
;;
  
let convert_pattern str =
  Array.init (String.length str) (fun i ->
    match String.get str i with 
      | '*' -> PLAYER
      | ' ' -> IGNORE
      | '#' -> IS_WALL_OR_WATER
      | '.' -> OPEN_NOT_BEACH
      | ':' -> MAYBE_BEACH
      | '^' -> ROBOT_N
      | '>' -> ROBOT_E
      | 'v' -> ROBOT_S
      | '<' -> ROBOT_W
      | 'x' -> ROBOT_STILL
      | 'I' -> ROBOT_ROCK
      | '~' -> IS_WATER
      | 'w' -> ISNT_WATER
      | c -> raise (Bad_char c)
  )
;;

let rec read_pattern_rec inp list =
  let str = getline inp in
  if str = "" || str = "END" then
    list
  else
    (* this reads the lines in and puts them together *reversed *)
    (* this is appropriate because we want them to look right visually *)
    (* in the file, but be stored in cartesian coordinates *)
    read_pattern_rec inp ((convert_pattern str)::list)
;;

let find_source pat =
  let src = ref None in
  for j = 0 to (Array.length pat)-1 do
    for i = 0 to (Array.length pat.(j))-1 do
      if pat.(j).(i) == PLAYER then
        src := Some(i,j)
    done
  done;
  !src
;;

(* read a pattern given the first line of it *)

let read_pattern inp str =
  let command = match str with
    | "N" -> Move(N)
    | "E" -> Move(E)
    | "S" -> Move(S)
    | "W" -> Move(W)
    | "H" -> Pick([])
    | _ -> raise Bad_input
  and pattern = read_pattern_rec inp [] in
  let pat_array = Array.of_list pattern in
  let src = find_source pat_array in
  match src with
    | Some(src_loc) ->
       { pat_action = command; pat = Array.of_list pattern; source = src_loc}
    | None -> print_endline "Doh!";raise Pattern_lacks_origin
;;

(* read the patterns recurisvely into a list *)
(* note this stores them backwards, so reverse later *)

let rec read_patterns inp list =
  let x = getline inp in
  if x = "END"
    then list
    else if x = ""
      then read_patterns inp list
      else read_patterns inp ((read_pattern inp x)::list)
;;

(* read the patterns into an array *)

let load_patterns name =
  let inp = open_in name in
  let list = read_patterns inp [] in
  (Array.of_list (List.rev list))     
